open Kernel

open Interpret

open Data.Type

let wait_done event k =
  let m = Mutex.create () in
  let c = Condition.create () in
  add_thread
    (fun () ->
       ask event k (Sgn.get "done" 0)
       (fun event tok ->
          Mutex.lock m;
          Condition.signal c;
          Mutex.unlock m
       ));
  Mutex.lock m;
  Condition.wait c m;
  Mutex.unlock m

let expat_parse event t =
  let s = string_proj t.(0) in
  let m = t.(1) in
  let parse_thread () =
    let p = Expat.parser_create_ns None ':' in
    let start_element_handler tag args =
      let tag = string_inj tag in
      let args =
        List.map
          (fun (a, b) -> array_ty.inj [| string_inj a; string_inj b |]) args in
      let args = list_inj args in
      let kv = V.fresh () in
      let k = var_ty.inj kv in
      let _ =
        tell event m (Compound.mk "start_element" [| tag; args; k |]) in
      wait_done event k in
    Expat.set_start_element_handler p start_element_handler;
    let end_element_handler tag =
      let tag = string_inj tag in
      let kv = V.fresh () in
      let k = var_ty.inj kv in
      let _ = tell event m (Compound.mk "end_element" [| tag; k |]) in
      wait_done event k in
    Expat.set_end_element_handler p end_element_handler;
    let character_data_handler data =
      let data = string_inj data in
      let kv = V.fresh () in
      let k = var_ty.inj kv in
      let _ =
        tell event m (Compound.mk "character_data" [| data; k |]) in
      wait_done event k in
    Expat.set_character_data_handler p character_data_handler;
    Expat.parse p s;
    ignore (tell event t.(2) (Compound.mk "done" [| |])) in
  add_thread parse_thread

let () = decl_simple_builtin (Sgn.get "expat_parse" 3) expat_parse

let dummy () = ()
